"""
BasePlayer, (c) 2022-present, animdl authors

Note from developer, KR:

This is the base class for all players. All players
**must** inherit from this class.

    class YourPlayer(BasePlayer):
        ...

The .play() method should not necessarily call .spawn().

If you can handle multiple streams with a single process,
that is suggested.

Your class must make use of .opts_spec to map options with
their corresponding arguments. This is done so that the
class may be re-used when writing similar player wrappers.

Your class must have a .optimisation_args attribute. May it
be an empty tuple. This is done so that any developer may
add forced arguments to the player for project-specific
optimisations.

Do not make .play() delay the execution of the player or wait
for the process to complete. You must make use of .process
attribute to do so.

.spawn() kills pre-existing processes. Do not change this behaviour.

.indicate_error() is a method that is called when a player
process has been deemed offline. If an error occured in the player,
during playback, override this method to indicate the error.

This feature is to make handling errors from the player easier
irrespective of the player persistence.

Lastly,

This is not made for any specific project. These players are made
so that they can be used in any project. Hence, do not make
anything with project-specific code.
"""

import subprocess


class BasePlayer:

    opts_spec: dict
    optimisation_args: tuple

    def __init__(self, executable, args=()):
        self.executable = executable
        self.args = args

        self.process: subprocess.Popen = None

    def play(self):
        raise NotImplementedError

    def spawn(self, *args, **kwargs):
        if self.process is not None:
            self.process.kill()

        self.process = subprocess.Popen(*args, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.process is not None:
            self.process.wait()

    def indicate_error(self):
        return self.process.returncode



from functools import partial

import lxml.html as htmlparser
import regex

from ....config import NINEANIME
from ...helpers import construct_site_based_regex
from .decipher import decrypt_url

CONTENT_ID_REGEX = regex.compile(r'data-id="(.+?)"')

REGEX = construct_site_based_regex(
    NINEANIME, extra_regex=r"/watch/[^&?/]+\.(?P<slug>[^&?/]+)"
)
TITLES_REGEX = regex.compile(r'<h1 class="title d-title" .+?>(.+?)</h1>')

SOURCES = {
    "41": "vidstream",
    "28": "mycloud",
    "35": "mp4upload",
    "40": "streamtape",
    "43": "videovard",
    "44": "filemoon",
}

DECRYPTION_CODE = "hlPeNwkncH0fq9so"


def fetch_episode(session, data_source):

    response = htmlparser.fromstring(
        session.get(
            NINEANIME + f"ajax/server/list/{data_source.get('data-ids')}"
        ).json()["result"]
    )

    for content_type_container in response.cssselect("div[data-type]"):

        for servers in content_type_container.cssselect("ul > li[data-sv-id]"):

            source = SOURCES[servers.get("data-sv-id")]

            content_url = decrypt_url(
                session.get(
                    NINEANIME + f"ajax/server/{servers.get('data-link-id')}/"
                ).json()["result"]["url"],
                DECRYPTION_CODE,
            )
            yield {
                "stream_url": content_url,
                "further_extraction": (
                    source,
                    {"headers": {"referer": NINEANIME}},
                ),
                "title": f"Episode {data_source.get('data-num')}, {content_type_container.get('data-type').upper()}",
            }


def fetcher(session, url, check, match):

    content_id = CONTENT_ID_REGEX.search(session.get(url).text).group(1)

    for data_source in htmlparser.fromstring(
        session.get(
            NINEANIME + f"ajax/episode/list/{content_id}",
        ).json()["result"]
    ).cssselect("a[data-num]"):
        yield partial(
            lambda data_source: list(fetch_episode(session, data_source)),
            data_source,
        ), data_source.get("data-num")


def metadata_fetcher(session, url, match):

    response = session.get(url).text

    return {"titles": TITLES_REGEX.findall(response)}



from warnings import warn

from .base_player import BasePlayer


class VLCPlayer(BasePlayer):

    optimisation_args = ("--http-forward-cookies",)

    opts_spec = {
        "http_referrer": "--http-referrer",
        "user_agent": "--user-agent",
        "subtitle": "--sub-file",
    }

    def play(
        self, stream_url, subtitles=None, headers=None, title=None, opts=None, **kwargs
    ):
        args = (self.executable, stream_url)

        if opts is not None:
            args += tuple(opts)

        if headers is not None:
            lowercased_headers = {key.lower(): value for key, value in headers.items()}

            keys = set(lowercased_headers.keys())

            if "user-agent" in keys:
                args += (
                    f"{self.opts_spec['user_agent']}={lowercased_headers['user-agent']}",
                )

            if "referer" in keys:
                args += (
                    f"{self.opts_spec['http_referrer']}={lowercased_headers['referer']}",
                )

            extra = keys - {"user-agent", "referer"}

            if extra:
                warn(
                    f"VLC does not support {', '.join(map(repr, extra))} headers. "
                    "This may result in stream loading failure."
                )

        if title is not None:
            args += (f"--meta-title={title}",)

        if subtitles is not None:
            args += tuple(f"--sub-file={subtitles}" for subtitles in subtitles)

        self.spawn(args)



from .android import AndroidIntentVIEW
from .ffplay import FFPlay
from .mpv import CelluloidPlayer, IINAPlayer, MPVDefaultPlayer
from .vlc import VLCPlayer

player_mapping = {
    "mpv": MPVDefaultPlayer,
    "iina": IINAPlayer,
    "vlc": VLCPlayer,
    "celluloid": CelluloidPlayer,
    "ffplay": FFPlay,
    "android": AndroidIntentVIEW,
}



from collections import defaultdict
from functools import partial

import yarl

from ....config import (
    ALLANIME,
    SUPERANIME_CRUNCHYROLL,
    SUPERANIME_RETURN_ALL,
    SUPERANIME_TYPE_OF,
)
from ...helpers import construct_site_based_regex, optopt
from .superscrapers import iter_unpacked_from_packed_hls

REGEX = construct_site_based_regex(ALLANIME, extra_regex=r"/anime/([^?&/]+)")

SOURCE_EMBED_REGEX = optopt.regexlib.compile(
    r'<iframe id="episode-frame" .+? src="(.+?)"'
)
SOURCE_URLS = optopt.regexlib.compile(
    r'sourceUrl[:=]"(?P<url>.+?)"[;,](?:.+?\.)?priority[:=](?P<priority>.+?)[;,](?:.+?\.)?sourceName[:=](?P<name>.+?)[,;]'
)
EPISODES_REGEX = optopt.regexlib.compile(r'\\"availableEpisodesDetail\\":({.+?})')
TITLES_REGEX = optopt.regexlib.compile(r'<span class="mr-1">(.+?);?</span>')

EMBED_RESOLVERS = {
    '"Ok"': "okru",
    '"Vid-mp4"': "gogoanime",
    '"Mp4"': "mp4upload",
    '"Sl-mp4"': "streamlare",
    '"Ss-Hls"': "streamsb",
}


def iter_episodes(
    episode_dictionary: dict,
    anime_page_url: str,
    *,
    selected_type_of: str = SUPERANIME_TYPE_OF,
):
    episodes = defaultdict(list)

    for type_of, episode_numbers in episode_dictionary.items():
        if type_of != selected_type_of:
            continue

        for episode in episode_numbers:
            episodes[int(episode) if episode.isdigit() else 0].append(
                anime_page_url + "/episodes/{}/{}".format(type_of, episode)
            )

    yield from sorted(episodes.items(), key=lambda x: x[0])


def unicode_escape(string: str):
    return string.encode("utf-8").decode("unicode_escape")


def to_json_url(url: "yarl.URL"):
    return url.with_name(url.name + ".json").with_query(url.query)


def iter_prioritised(session, urls):

    for url, (priority, name) in urls:
        data = session.get(url.human_repr()).text

        if data == "Wrongerror":
            continue

        json_parsed = optopt.jsonlib.loads(data)["links"]

        def iter_all():
            for link in json_parsed:

                stream_attr = {}

                if "resolution" in link:
                    stream_attr["quality"] = link["resolution"]

                if "subtitles" in link:
                    stream_attr["subtitle"] = [_["src"] for _ in link["subtitles"]]

                if "resolutionStr" in link:
                    stream_attr["vrv"] = {
                        "stream_type": link["resolutionStr"],
                        "provider_configuration": SUPERANIME_CRUNCHYROLL,
                    }

                stream_url = link["link"]

                yield from (
                    iter_unpacked_from_packed_hls(
                        session, yarl.URL(stream_url), stream_attribs=stream_attr
                    )
                )

        yield list(iter_all())


def extract_content(
    session,
    content: "iter_episodes",
    *,
    api_endpoint: "yarl.URL",
    return_all: bool = SUPERANIME_RETURN_ALL,
):

    for url in content:

        direct_providers = set()
        embed_providers = set()

        content_page = session.get(url).text

        has_on_embed = SOURCE_EMBED_REGEX.search(content_page)

        if has_on_embed:
            direct_providers.add(
                (to_json_url(yarl.URL(has_on_embed.group(1))), (0, "embed"))
            )
        else:
            for source_urls in SOURCE_URLS.finditer(content_page):

                raw_url = unicode_escape(source_urls.group(1))
                parsed_url = yarl.URL(raw_url)

                priority, name = source_urls.group("priority", "name")

                if parsed_url.host is None:
                    parsed_url = api_endpoint.join(to_json_url(parsed_url))
                    direct_providers.add((parsed_url, (priority, name)))
                else:
                    embed_providers.add((parsed_url, (priority, name)))

        iterator = iter_prioritised(
            session, sorted(direct_providers, key=lambda x: x[1][0])
        )

        if return_all:
            yield from iterator
        else:
            child = next(iterator, None)

            if child is not None:
                yield from child


def fetcher(session, url: "str", check, match):
    anime_url = ALLANIME + f"anime/{match.group(1)}"

    api_endpoint = yarl.URL(
        session.get(ALLANIME + "getVersion").json().get("episodeIframeHead", "")
    )

    for episode, content in iter_episodes(
        optopt.jsonlib.loads(
            EPISODES_REGEX.search(session.get(anime_url).text)
            .group(1)
            .replace('\\"', '"')
        ),
        anime_url,
    ):
        if check(episode):
            yield partial(
                lambda session, content: list(
                    extract_content(session, content, api_endpoint=api_endpoint)
                ),
                session,
                content,
            ), episode


def metadata_fetcher(session, url: "str", match):
    anime_url = ALLANIME + f"anime/{match.group(1)}"

    return {"titles": TITLES_REGEX.findall(session.get(anime_url).text)}



import logging

import click

from ...__version__ import __core__
from ...codebase import providers
from ...config import (
    CHECK_FOR_UPDATES,
    DEFAULT_PLAYER,
    DEFAULT_PROVIDER,
    DISCORD_PRESENCE,
    FORCE_STREAMING_QUALITY_SELECTION,
    QUALITY,
    USE_ANISKIP,
)
from .. import exit_codes, helpers, http_client

if DISCORD_PRESENCE:
    try:
        from ..helpers.rpc import set_streaming_episode
    except ImportError:
        raise ImportError(
            "Discord RPC was set to be enabled but `pypresence` is not installed, install it with `pip install pypresence`."
        )
    except RuntimeError:
        DISCORD_PRESENCE = False


@click.command(name="stream", help="Stream your favorite anime by query.")
@helpers.decorators.content_fetch_options(
    default_quality_string=QUALITY,
)
@helpers.decorators.player_options(default_player=DEFAULT_PLAYER)
@helpers.decorators.automatic_selection_options()
@helpers.decorators.logging_options()
@helpers.decorators.setup_loggers()
@helpers.decorators.banner_gift_wrapper(
    http_client.client, __core__, check_for_updates=CHECK_FOR_UPDATES
)
def animdl_stream(
    query, special, quality, player_opts, player, index, log_level, **kwargs
):
    """
    Streamer call for animdl streaming session.
    """
    r = kwargs.get("range")

    logger = logging.getLogger("streamer")

    if player_opts is not None:
        player_opts = tuple(click.parser.split_arg_string(player_opts))
    else:
        player_opts = ()

    try:
        streamer = helpers.player.handle_player(DEFAULT_PLAYER, player_opts, player)
    except Exception as e:
        logger.error(
            "Streaming failed due to: {!r}.".format(e),
            exc_info=True,
        )
        raise SystemExit(exit_codes.STREAMER_CONFIGURATION_REQUIRED)

    anime, provider = helpers.process_query(
        http_client.client, query, logger, auto_index=index, provider=DEFAULT_PROVIDER
    )

    if not anime:
        logger.critical("Searcher returned no anime to stream, failed to stream.")
        raise SystemExit(exit_codes.NO_CONTENT_FOUND)

    logger.name = "{}/{}".format(provider, logger.name)
    logger.debug("Will scrape from {}".format(anime))
    logger.info("Now initiating your stream session")

    match, provider_module, _ = providers.get_provider(
        providers.append_protocol(anime.get("anime_url"))
    )

    streams = list(
        provider_module.fetcher(http_client.client, anime.get("anime_url"), r, match)
    )

    if special:
        streams = list(helpers.special.special_parser(streams, special))

    content_title = anime["name"]

    total = len(streams)

    if total < 1:
        logger.critical("No streams found, failed to stream.")
        raise SystemExit(exit_codes.NO_CONTENT_FOUND)

    for count, (stream_urls_caller, episode_number) in enumerate(streams, 1):

        if episode_number == 0:
            episode_text = "Special Episode"
        else:
            episode_text = f"Episode {episode_number}"

        if content_title:
            media_title = f"{content_title}: {episode_text}"
        else:
            media_title = episode_text

        playing = True

        while playing:

            stream_urls = list(
                helpers.ensure_extraction(http_client.client, stream_urls_caller)
            )

            if not stream_urls:

                logger.warning(
                    f"Could not find stream urls for {content_title!r}. "
                    "The episode may not be available or the scraper has broke."
                )

                if log_level <= logging.DEBUG:
                    playing = click.confirm("Retry stream url extraction?")

                continue

            titles = set(_.get("title") for _ in stream_urls)

            if len(titles) > 1:
                logger.warning(
                    f"Multiple titles found for the same streams, multiple seasons may be available. Please adjust the quality string as required: {', '.join(map(repr, titles))}"
                )
                logger.info(
                    "quality_string set to \"best[title='(Season XYZ)']\" may work. (use r'.+?' for regex support!)"
                )

            selection = helpers.prompts.quality_prompt(
                logger,
                log_level,
                stream_urls,
                force_selection_string=quality
                if FORCE_STREAMING_QUALITY_SELECTION
                else None,
            )

            headers = selection.get("headers", {})

            logger.info(
                f"Streaming {media_title!r}, [{count:d}/{total:d}, {total - count:d} remaining]"
            )
            logger.debug(f"Obtained streams: {stream_urls!r}")

            if "title" in selection:
                media_title += " ({})".format(selection["title"])

            if USE_ANISKIP:
                chapters = helpers.aniskip.get_timestamps(
                    http_client.client, content_title, episode_number
                )
            else:
                chapters = []

            with streamer:
                streamer.play(
                    selection["stream_url"],
                    title=media_title,
                    headers=headers,
                    subtitles=selection.get("subtitle", []),
                    chapters=chapters,
                )

                if DISCORD_PRESENCE:
                    set_streaming_episode(
                        http_client.client, content_title, episode_number
                    )

            error_indication = streamer.indicate_error()

            if bool(error_indication):
                logger.warning(f"Streamer indicated an error {error_indication!r}.")

                if log_level <= logging.INFO:
                    if click.confirm("Retry streaming this episode?"):
                        continue

            if log_level <= logging.DEBUG:
                if click.confirm("Replay episode?"):
                    continue

            playing = False



"""
Internet Download Manager Support.

This is made strictly for Windows.
"""

import time

IDM_MID = ["{ECF21EAB-3AA8-4355-82BE-F777990001DD}", 1, 0]
DOWNLOAD_FOLDER = "Downloads"

idmlib, client = None, None


def supported():
    return bool(idmlib)


try:
    import comtypes.client as cc

    idmlib = cc.GetModule(IDM_MID)

    client = cc.CreateObject(
        idmlib.CIDMLinkTransmitter, None, None, idmlib.ICIDMLinkTransmitter2
    )
except BaseException:
    pass


def within_range(t, t1, t2):
    return t1 <= t <= t2


def idm_download(
    url,
    headers={},
    form_data="",
    auth=(None, None),
    filename="",
    download_folder=DOWNLOAD_FOLDER,
    lflag=5,
):
    return client.SendLinkToIDM(
        url,
        headers.get("referer", ""),
        headers.get("cookie", ""),
        form_data,
        auth[0] or "",
        auth[1] or "",
        download_folder,
        filename,
        lflag,
    )


def wait_until_download(
    url,
    headers={},
    form_data="",
    auth=(None, None),
    filename="",
    download_folder=DOWNLOAD_FOLDER,
):
    idm_download(
        url, headers, form_data, auth, filename, download_folder.as_posix(), lflag=5
    )
    while not (download_folder / filename).exists():
        try:
            time.sleep(0.5)
        except KeyboardInterrupt:
            return print(
                "[IDM] Interrupted the wait for current download completion. Continuing with the next in queue."
            )



"""
Special arguments parser
"""

import regex

SPECIAL_PARSER = regex.compile(r"(?P<special>\D+)(-?\d*)", regex.IGNORECASE)

strings = {
    "last": ["l", "latest"],
}


def get_qualified_name(special):
    for key, values in strings.items():
        if special.lower() in (*values, key):
            return key


def special_parser(streams, string):

    returnee = []

    for match in SPECIAL_PARSER.finditer(string):
        special, index = match.group("special", 2)
        index = None if not index else int(index)

        special = get_qualified_name(special)

        if special is None:
            continue

        if special == "last":
            returnee.extend(streams[-(index or 1) :])
        else:
            """
            More specials to be thought about.
            """

    yield from returnee
    yield from (stream for stream in streams if stream not in returnee)



import lxml.html as htmlparser

from ....config import HAHO
from ...helpers import construct_site_based_regex
from ..tenshimoe import fetcher as tenshimoe_fetcher
from ..tenshimoe import metadata_fetcher as tenshimoe_metadata_fetcher

REGEX = construct_site_based_regex(HAHO, extra_regex=r"/anime/([^?&/]+)")


def post_processor(session, stream_page):
    def _get_quality(quality_string):
        q = quality_string.rstrip("pP")
        if q.isdigit():
            return int(q)
        return None

    yield from (
        {"stream_url": _.get("src"), "quality": _get_quality(_.get("title"))}
        for _ in htmlparser.fromstring(stream_page).cssselect("source")
    )


def fetcher(*args, **kwargs):
    yield from tenshimoe_fetcher(
        *args, **kwargs, post_processor=post_processor, domain=HAHO
    )


def metadata_fetcher(*args, **kwargs):
    return tenshimoe_metadata_fetcher(*args, **kwargs, domain=HAHO)



from .base_player import BasePlayer


class FFPlay(BasePlayer):

    opts_spec = {
        "headers": "-headers",
        "metadata": "-metadata",
        "title_field": "title",
    }

    def play(self, stream_url, headers=None, title=None, opts=None, **kwargs):
        args = (self.executable, stream_url)

        if opts is not None:
            args += tuple(opts)

        if headers is not None:
            args += (
                f"{self.opts_spec['headers']}={self.headers_joiner.join(f'{key}: {value}' for key, value in headers.items())}",
            )

        if title is not None:
            args += (self.opts_spec["metadata"], f"{self.opts_spec['title']}={title}")

        self.spawn(args)



from typing import Callable, Iterable, TypeVar

import regex

search_type = TypeVar("search_type")

options = {
    "ignore_whitespace": lambda x: regex.sub(r"\s+", "", x),
    "ignore_punctuation": lambda x: regex.sub(r"\p{P}+", "", x),
}


def search(
    query: "str",
    possibilities: "Iterable[search_type]",
    *,
    processor: "Callable[[search_type], str]" = lambda r: r,
    search_options: tuple = ("ignore_punctuation",),
):

    pattern = regex.compile(
        r"(.*?)".join(map(regex.escape, query.strip())),
        flags=regex.IGNORECASE,
    )

    def genexp():
        for search_value in possibilities:
            processed_search_value = processor(search_value)

            for option in search_options:
                if option in options:
                    processed_search_value = options[option](processed_search_value)

            match = pattern.search(processed_search_value)
            if match:
                yield len(processed_search_value) - len(query.strip()), search_value

    for _, search_value in sorted(genexp(), key=lambda x: x[0]):
        yield search_value



import regex

from ....config import ANIMIXPLAY
from ...helpers import construct_site_based_regex
from .stream_url import fetcher

REGEX = construct_site_based_regex(ANIMIXPLAY, extra_regex=r"/v\d+/([^?&/]+)")

TITLES_REGEX = regex.compile(r'<span class="animetitle">(.+?)</span>')


def metadata_fetcher(session, url, match):
    return {"titles": TITLES_REGEX.findall(session.get(url).text)}



import operator
import platform
import re
import sys
from functools import lru_cache

import pkginfo
from packaging import version

python_impl_dict = {
    "CPython": "cp",
    "PyPy": "pp",
    "Jython": "jy",
    "IronPython": "ip",
    "Stackless": "sl",
    "Brython": "br",
}


PYTHON_IMPLEMENTATION = python_impl_dict.get(platform.python_implementation())

if PYTHON_IMPLEMENTATION is None:
    raise ValueError(
        f"Unsupported Python implementation {platform.python_implementation()!r}, will not find wheels."
    )


PACKAGE_NAME = "lxml"
VERSION_EXPR = "~=4.9.1"

PC_ARCH = "win_amd64" if sys.maxsize > 2**32 else "win32"
STRING_REGEX = re.compile(r"&#(\d+);")
ABI_TAG = f"{PYTHON_IMPLEMENTATION}{sys.version_info.major}{sys.version_info.minor}"


def match_version(version_expression: str, version_string: str) -> bool:
    if version_expression is None:
        return True

    def single_matcher(single_expression: str):

        target_version = single_expression.lstrip("=<>~")
        expr = single_expression[: -len(target_version)]

        target = version.parse(target_version)
        current = version.parse(version_string)

        operators = {
            "==": operator.eq,
            ">=": operator.ge,
            "<=": operator.le,
            ">": operator.gt,
            "<": operator.lt,
            "~=": lambda x, y: (x.major == y.major and x.minor == y.minor) or x >= y,
        }

        user_operator = operators.get(expr)

        if user_operator is None:
            raise ValueError(f"Invalid version expression {expr!r}")

        return user_operator(current, target)

    return all(
        single_matcher(single_expression)
        for single_expression in version_expression.split(",")
    )


@lru_cache()
def get_packages_page(session):
    return session.get(
        "https://www.lfd.uci.edu/~gohlke/pythonlibs/", follow_redirects=True
    ).text


def iter_packages(
    session, package_name, version_expression: str = None, pc_arch: str = PC_ARCH
):

    WHEELS_REGEX = re.compile(rf"]'>({package_name}.+?{pc_arch}&#46;whl)<")

    for _ in WHEELS_REGEX.finditer(get_packages_page(session)):
        package = STRING_REGEX.sub(lambda match: chr(int(match.group(1))), _.group(1))

        _, package_version, *abis, _ = package.split("‑")

        if match_version(version_expression, package_version) and ABI_TAG in abis:
            yield "https://download.lfd.uci.edu/pythonlibs/archived/" + package.replace(
                "‑", "-"
            )


def resolve(packages, *, prefix="[animdl/utils/win32-build-inquirer]"):
    import subprocess

    import httpx

    pip_install_args = [sys.executable, "-m", "pip", "install"]

    session = httpx.Client()

    for package, version_expr in packages:

        pkg = pkginfo.get_metadata(package)

        if pkg is not None and match_version(version_expr, pkg.version):
            continue

        pip_pkg_string = package + (version_expr if version_expr else "")

        print(
            prefix, f"{pip_pkg_string!r} is not installed, trying to install from PyPI."
        )

        pip_process = subprocess.Popen(
            pip_install_args + [pip_pkg_string],
        )
        pip_process.wait()

        if pip_process.returncode == 0:
            continue

        print(
            prefix,
            f"Failed to install {pip_pkg_string!r} from PyPI, trying to install from the build inquiry.",
        )

        if sys.platform != "win32":
            raise RuntimeError(
                f"Failed to install {pip_pkg_string!r} with pip. Build inquiry is not supported on {sys.platform!r}."
            )

        urls = list(iter_packages(session, package, version_expr))

        if not urls:
            raise RuntimeError(
                f"Failed to find a wheel for {pip_pkg_string!r} in the build inquiry."
            )

        selection = urls[0]

        print(prefix, f"Installing {pip_pkg_string!r} from {selection!r}.")

        pip_process = subprocess.Popen(
            pip_install_args + [selection],
        )
        pip_process.wait()

        if pip_process.returncode != 0:
            raise RuntimeError(
                f"Failed to install {pip_pkg_string!r} from the build inquiry, please raise an issue if this is persistent"
            )



from functools import partial

import lxml.html as htmlparser
import regex

from ....config import HENTAISTREAM
from ...helpers import construct_site_based_regex

EPISODE_REGEX = regex.compile(r"/\d+/[^&?/]+")
TITLES_REGEX = regex.compile(r'<h1 class="entry-title" itemprop="name">(.+?)</h1>')


REGEX = construct_site_based_regex(
    HENTAISTREAM, extra_regex=r"/hentai/([^?&/]+)(?:/\d+)?"
)

EXTRACTOR_REGEX = regex.compile(r"src: '(.+?)',\s+type: '.+?',\s+size: (\d+),")
SUBTITLE_URL_REGEX = regex.compile(r"subUrl: '(.+?)'")


def fetch_streams(session, episode_url, **opts):

    episode_page = session.get(HENTAISTREAM + episode_url[1:]).text
    subtitles = SUBTITLE_URL_REGEX.findall(episode_page)

    for url, size in EXTRACTOR_REGEX.findall(episode_page):

        yield {
            "stream_url": url,
            "quality": int(size),
            "headers": {"referer": HENTAISTREAM},
            "subtitle": subtitles,
        }


def fetcher(session, url, check, match):

    episodes_page = htmlparser.fromstring(
        session.get(HENTAISTREAM + f"hentai/" + match.group(1)).text
    )

    for urls in episodes_page.cssselect("div.eplister a"):
        number, title, date = map(
            htmlparser.HtmlElement.text_content, urls.cssselect('div[class^="epl"]')
        )

        episode_number = int(number) if number.isdigit() else 0

        if check(episode_number):
            yield partial(
                lambda url, **opts: list(fetch_streams(session, url, **opts)),
                urls.get("href"),
                title=f"{title} ({date})",
            ), episode_number


def metadata_fetcher(session, url, match):

    episode_url = HENTAISTREAM + f"hentai/" + match.group(1)

    return {"titles": TITLES_REGEX.findall(session.get(episode_url).text)}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['animdl',
 'animdl.core',
 'animdl.core.cli',
 'animdl.core.cli.commands',
 'animdl.core.cli.helpers',
 'animdl.core.cli.helpers.players',
 'animdl.core.codebase',
 'animdl.core.codebase.downloader',
 'animdl.core.codebase.extractors',
 'animdl.core.codebase.extractors.dailymotion',
 'animdl.core.codebase.extractors.doodstream',
 'animdl.core.codebase.extractors.gogoplay',
 'animdl.core.codebase.extractors.mp4upload',
 'animdl.core.codebase.extractors.mycloud',
 'animdl.core.codebase.extractors.okru',
 'animdl.core.codebase.extractors.rapidvideo',
 'animdl.core.codebase.extractors.streamlare',
 'animdl.core.codebase.extractors.streamsb',
 'animdl.core.codebase.extractors.streamtape',
 'animdl.core.codebase.extractors.videobin',
 'animdl.core.codebase.extractors.vidstream',
 'animdl.core.codebase.helpers',
 'animdl.core.codebase.providers',
 'animdl.core.codebase.providers.allanime',
 'animdl.core.codebase.providers.animekaizoku',
 'animdl.core.codebase.providers.animeonsen',
 'animdl.core.codebase.providers.animeout',
 'animdl.core.codebase.providers.animepahe',
 'animdl.core.codebase.providers.animepahe.inner',
 'animdl.core.codebase.providers.animexin',
 'animdl.core.codebase.providers.animixplay',
 'animdl.core.codebase.providers.animtime',
 'animdl.core.codebase.providers.crunchyroll',
 'animdl.core.codebase.providers.gogoanime',
 'animdl.core.codebase.providers.hahomoe',
 'animdl.core.codebase.providers.hentaistream',
 'animdl.core.codebase.providers.kamyroll',
 'animdl.core.codebase.providers.kawaiifu',
 'animdl.core.codebase.providers.nineanime',
 'animdl.core.codebase.providers.tenshimoe',
 'animdl.core.codebase.providers.twistmoe',
 'animdl.core.codebase.providers.zoro',
 'animdl.core.config']

package_data = \
{'': ['*']}

install_requires = \
['anchor-kr>=0.1.3,<0.2.0',
 'anitopy>=2.1.0,<2.2.0',
 'click>=8.0.4,<8.1.0',
 'comtypes>=1.1.11,<1.2.0',
 'cssselect>=1.1.0,<1.2.0',
 'httpx>=0.23.0,<0.24.0',
 'packaging>=22.0,<23.0',
 'pkginfo>=1.9.2,<2.0.0',
 'pycryptodomex>=3.14.1,<3.15.0',
 'pyyaml>=6.0,<7.0',
 'regex>=2022.10.31,<2022.11.0',
 'tqdm>=4.62.3,<4.63.0',
 'yarl>=1.8.1,<1.9.0']

entry_points = \
{'console_scripts': ['animdl = animdl.__main__:__animdl_cli__']}

setup_kwargs = {
    'name': 'animdl',
    'version': '1.6.50',
    'description': 'A highly efficient, fast, powerful and light-weight anime downloader and streamer for your favorite anime.',
    'long_description': '',
    'author': 'justfoolingaround',
    'author_email': 'kr.justfoolingaround@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from collections import defaultdict
from functools import lru_cache, partial

from ....config import CRUNCHYROLL, SUPERANIME_CRUNCHYROLL
from ... import helpers
from .api import Kamyroll

VRV_RESPONSE_REGEX = helpers.optopt.regexlib.compile(
    r"^#EXT-X-STREAM-INF:.*?RESOLUTION=\d+x(?P<resolution>\d+).*?\n(.+?)$",
    flags=helpers.optopt.regexlib.MULTILINE,
)
CR_GUID_REGEX = helpers.optopt.regexlib.compile(r"crunchyroll\.com/.+?/(.+?)/.+?")


REGEX = helpers.optopt.regexlib.compile(
    r"kamyroll://(?:.+?\.)+?crunchyroll\.com/([^&?/]+)"
)


def fetch_streams(api: Kamyroll, medias: "list"):

    superanime_subtitle = SUPERANIME_CRUNCHYROLL.get("subtitle_language") or ""

    for stream_attrs, media_id in medias:
        streams_data = api.fetch_streams(media_id)

        if not streams_data["streams"]:
            continue

        selected_streams = [
            _
            for _ in streams_data["streams"]
            if _["hardsub_locale"] == superanime_subtitle
        ] or streams_data["streams"]
        subtitles = streams_data["subtitles"]

        for stream in selected_streams:
            hls_response = api.session.get(stream["url"]).text

            for match in VRV_RESPONSE_REGEX.finditer(hls_response):
                stream_url = match.group(2).replace("/index-v1-a1.m3u8", "")

                yield {
                    "stream_url": stream_url,
                    "quality": int(match.group(1)),
                    "subtitle": [_["url"] for _ in subtitles],
                    **stream_attrs,
                }


@lru_cache()
def fetch_episodes(api: Kamyroll, media_id):
    seasons = api.fetch_seasons(media_id)["items"]
    episodes = defaultdict(list)

    for season in seasons:
        for episode in season["episodes"]:
            episode_number = episode["episode_number"] or 0
            episodes[episode_number].append(
                (
                    {
                        "title": f'{episode["title"]} ({season["title"]})',
                    },
                    episode["id"],
                )
            )
    return episodes


@lru_cache()
def get_crunchyroll_guid(session, url):
    cr_guid = CR_GUID_REGEX.search(
        session.get(
            CRUNCHYROLL + url,
            headers={"Referer": "https://google.com/"},
            follow_redirects=False,
        ).headers.get("Location")
    )

    if cr_guid is None:
        return

    return cr_guid.group(1)


def fetcher(session, url, check, match):
    api = Kamyroll(session)

    cr_guid = get_crunchyroll_guid(session, match.group(1))

    if cr_guid is None:
        return

    episodes = fetch_episodes(api, cr_guid)

    sorted_episodes = sorted(episodes.items(), key=lambda x: x[0])

    for episode_number, medias in sorted_episodes:
        if check(episode_number):
            yield partial(
                lambda medias: list(fetch_streams(api, medias)),
                medias,
            ), episode_number


def metadata_fetcher(session, url, match):
    api = Kamyroll(session)

    cr_guid = get_crunchyroll_guid(session, match.group(1))

    if cr_guid is None:
        return

    metadata = api.fetch_media(cr_guid)

    return {
        "title": [metadata["title"]],
    }



from functools import partial

import yarl

from ....config import TWIST
from ...helpers import construct_site_based_regex, parse_from_content
from .stream_url import *

REGEX = construct_site_based_regex(TWIST, extra_regex=r"/a/([^?&/]+)")


def fetcher(session, url, check, match):
    anime_name = match.group(1)

    for episode, stream in sorted(
        iter_episodes(session, anime_name), key=lambda k: k[0]
    ):
        if check(episode):
            yield partial(
                lambda s: [
                    parse_from_content(
                        yarl.URL(s),
                        name_processor=lambda u: u.name,
                        stream_url_processor=lambda u: u.human_repr(),
                        overrides={"headers": {"referer": "https://twist.moe/"}},
                        episode_parsed=True,
                    )
                ],
                stream,
            ), episode


def metadata_fetcher(session, url, match):
    return {
        "titles": [
            session.get(
                "https://api.twist.moe/api/anime/{}".format(match.group(1)),
                headers={"x-access-token": "0df14814b9e590a1f26d3071a4ed7974"},
            )
            .json()
            .get("title")
        ]
    }
